%%%----------------------------------------------------------------------
%%% File    : cmdline_if.erl
%%% Author  : Tomas Abrahamsson <tab@lysator.liu.se>
%%% Purpose : Command line interface
%%% Created : 30 May 2000 by Tomas Abrahamsson <tab@lysator.liu.se>
%%%----------------------------------------------------------------------

%%%----------------------------------------------------------------------
%%% trelay -- a program for relaying tcp/udp traffic
%%% Copyright (C) 2000  Tomas Abrahamsson <tab@lysator.liu.se>
%%%
%%% This program is free software; you can redistribute it and/or modify
%%% it under the terms of the GNU General Public License as published by
%%% the Free Software Foundation; either version 2 of the License, or
%%% (at your option) any later version.
%%%
%%% This program is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU General Public License for more details.
%%%
%%% You should have received a copy of the GNU General Public License
%%% along with this program; if not, write to the Free Software
%%% Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
%%%----------------------------------------------------------------------


-module(cmdline_if).
-author('tab@lysator.liu.se').
-vsn('$Id$').   % '


%%-compile(export_all).
%%-export([Function/Arity, ...]).

-export([start/1, start_link/1]).

%% exports due to spawns:
-export([cmdlineserver_init/2, input_init/1]).

-record(state, {halt_on_eof = true,
                has_seen_eof = false}).

start(RelayMgr) ->
    spawn(?MODULE, cmdlineserver_init, [RelayMgr, self()]).

start_link(RelayMgr) ->
    spawn_link(?MODULE, cmdlineserver_init, [RelayMgr, self()]).

cmdlineserver_init(RelayMgr, From) ->
    process_flag(trap_exit, true),
    InputServer = input_start(),
    print_welcome_msg(),
    InputServer ! start,
    cmdlineserver_loop(RelayMgr, From, #state{}).

cmdlineserver_loop(RelayMgr, From, State) ->
    receive
        {line, LineInputServer, Line} ->
            ParsedLine = parse_line(Line),
            case eval_line(ParsedLine, RelayMgr, State) of
                {continue, NewState} ->
                    case State#state.has_seen_eof of
                        false ->
                            LineInputServer ! continue;
                        true ->
                            do_nothing
                    end,
                    cmdlineserver_loop(RelayMgr, From, NewState);
                {stop, _NewState} ->
                    LineInputServer ! stop,
                    From ! done
            end;
        {'EXIT', Pid, Reason} ->
            io:format("Exit message from ~p: ~p~n", [Pid, Reason]),
            cmdlineserver_loop(RelayMgr, From, State)
    end.


parse_line(eof) ->
    eof;
parse_line(Line) ->
    StrippedLine = string:strip(Line),
    parse_words(string:tokens(StrippedLine, " "), StrippedLine).

parse_words(["start", RelayDescr], _OrigLine) ->
    case string:tokens(RelayDescr, "-") of
        [ProtoText, FromText, ToText] ->
            ProtoRes = parse_proto(ProtoText),
            FromAddrRes = parse_addr(FromText),
            ToAddrRes = parse_addr(ToText),
            case analyze([ProtoRes, FromAddrRes, ToAddrRes]) of
                {error, Error} ->
                    {error, start, Error};
                {ok, [Proto, FromAddr, ToAddr]} ->
                    {start, Proto, FromAddr, ToAddr, []}
            end;
        _Other ->
            {error, start, "badly formed expression"}
    end;
parse_words(["start", RelayDescr | OptionsTexts], _OrigLine) ->
    case string:tokens(RelayDescr, "-") of
        [ProtoText, FromText, ToText] ->
            ProtoRes = parse_proto(ProtoText),
            FromAddrRes = parse_addr(FromText),
            ToAddrRes = parse_addr(ToText),
            OptionsRes = parse_options(OptionsTexts),
            case analyze([ProtoRes, FromAddrRes, ToAddrRes, OptionsRes]) of
                {error, Error} ->
                    {error, start, Error};
                {ok, [Proto, FromAddr, ToAddr, Options]} ->
                    {start, Proto, FromAddr, ToAddr, Options};
                _Other ->
                    {error, start, "Parseerror"}
            end;
        _Other ->
            {error, start, "badly formed expression"}
    end;
parse_words(["stop", RelayDescrOrRef], _OrigLine) ->
    case catch list_to_integer(RelayDescrOrRef) of
        Ref when is_integer(Ref) ->
            {stop, Ref};
        _ConversionError ->
            case string:tokens(RelayDescrOrRef, "-") of
                [ProtoText, FromText, ToText] ->
                    ProtoRes = parse_proto(ProtoText),
                    FromAddrRes = parse_addr(FromText),
                    ToAddrRes = parse_addr(ToText),
                    case analyze([ProtoRes, FromAddrRes, ToAddrRes]) of
                        {error, Error} ->
                            {error, stop, Error};
                        {ok, [Proto, FromAddr, ToAddr]} ->
                            {stop, Proto, FromAddr, ToAddr}
                    end;
                _Other ->
                    {error, stop, "badly formed expression"}
            end
    end;
parse_words(["stop", RelayRefText, ConnRefText], _OrigLine) ->
    RelayRef = (catch list_to_integer(RelayRefText)),
    ConnRef = (catch list_to_integer(ConnRefText)),
    if
        is_integer(RelayRef), is_integer(ConnRef) ->
            {stop, RelayRef, ConnRef};
        true ->
            {error, stop, "bad reference"}
    end;
parse_words(["stopall"], _OrigLine) ->
    stopall;
parse_words(["changeopts", RelayRefText | OptionsText], _OrigLine) ->
    RelayRef = (catch list_to_integer(RelayRefText)),
    if
        is_integer(RelayRef) ->
            case parse_options(OptionsText) of
                {ok, Options} ->
                    {change_options, RelayRef, Options};
                {error, _Error} ->
                    {error, changeoptions, "bad options"}
            end;
        true ->
            {error, changeoptions, "bad relay reference"}
    end;
parse_words(["list"], _OrigLine) ->
    list;
parse_words(["save", FileName], _OrigLine) ->
    {save, FileName};
parse_words(["load", FileName], _OrigLine) ->
    {load, FileName};
parse_words(["erlang" | _Rest], OrigLine) ->
    ErlangStartIndex = string:str(OrigLine, "erlang"),
    ErlText = string:substr(OrigLine, ErlangStartIndex + string:len("erlang")),
    case parse_term(ErlText) of
        {ok, {Module, Function, Args}} ->
            {erlang, {Module, Function, Args}};
        {ok, _OtherTerm} ->
            {error, erlang, "bad term"};
        {error, _ErrInfo} ->
            {error, erlang, "parse error"}
    end;
parse_words(["donthaltoneof"], _OrigLine) ->
    dont_halt_on_eof;
parse_words(["dont", "halt", "on", "eof"], _OrigLine) ->
    dont_halt_on_eof;
parse_words(["halt"], _OrigLine) ->
    halt;
parse_words(["quit"], _OrigLine) ->
    halt;
parse_words(["help"], _OrigLine) ->
    help;
parse_words(["?"], _OrigLine) ->
    help;
parse_words(["copying"], _OrigLine) ->
    showcopying;
parse_words(["show", "copying"], _OrigLine) ->
    showcopying;
parse_words(["warranty"], _OrigLine) ->
    showwarranty;
parse_words(["show", "warranty"], _OrigLine) ->
    showwarranty;
parse_words("", _OrigLine) ->
    empty;
parse_words([[$# | _Rest1] | _Rest2], _OrigLine) ->
    comment;
parse_words(_Other, OrigLine) ->
    {error, unknown, "Command not recognized: \"" ++ OrigLine ++ "\""}.



parse_proto("tcp") ->
    {ok, tcp};
parse_proto("udp") ->
    {ok, udp};
parse_proto(BadProtoText) ->
    {error, "Badly formed protocol: \"" ++ BadProtoText ++ "\""}.

parse_addr(AddrText) ->
    case string:tokens(AddrText, ":") of
        ["*", Port] ->
            {ok, {any, list_to_integer(Port)}};
        ["=", Port] ->
            {ok, {same, list_to_integer(Port)}};
        [Host, Port] ->
            {ok, {Host, list_to_integer(Port)}};
        _Other ->
            {error, "Bad address: \"" ++ AddrText ++ "\""}
    end.

parse_options(Options) ->
    analyze(lists:map(fun(Option) -> parse_option(Option) end, Options)).


parse_option(Option) ->
    parse_term(Option).


parse_term(KeyText) ->
    case list_to_term(KeyText) of
        {ok, Key} ->
            {ok, Key};
        {error, _ErrInfo} ->
            {error, "Badly formed term"}
    end.




analyze(Results) ->
    lists:foldl(fun({error, ErrTxt}, {error, AccErrTxt}) ->
                        {error, AccErrTxt ++ "\n" ++ ErrTxt};
                   ({error, ErrText}, {ok, _AccData}) ->
                        {error, ErrText};
                   ({ok, _Data}, {error, ErrText}) ->
                        {error, ErrText};
                   ({ok, Data}, {ok, AccData}) ->
                        {ok, AccData ++ [Data]}
                end,
                {ok, []},
                Results).


eval_line(eof, _RelayMgr, State) ->
    case State#state.halt_on_eof of
        true ->
            io:format("~nExiting.~n"),
            {stop, State#state{has_seen_eof = true}};
        false ->
            {continue, State#state{has_seen_eof = true}}
    end;
eval_line(halt, _RelayMgr, State) ->
    io:format("Exiting.~n"),
    {stop, State};
eval_line(showcopying, _RelayMgr, State) ->
    show_legal:show_copying(),
    {continue, State};
eval_line(showwarranty, _RelayMgr, State) ->
    show_legal:show_warranty(),
    {continue, State};
eval_line(empty, _RelayMgr, State) ->
    {continue, State};
eval_line(comment, _RelayMgr, State) ->
    {continue, State};
eval_line({error, unknown, ErrText}, _RelayMgr, State) ->
    io:format("Error: ~s~n", [ErrText]),
    {continue, State};
eval_line({error, IdentifiedCommand, ErrText}, _RelayMgr, State) ->
    io:format("~p Error: ~s~n", [IdentifiedCommand, ErrText]),
    {continue, State};
eval_line({start, Proto, From, To, Options}, _RelayMgr, State) ->
    case trelay_mgr:start_relay(Proto, From, To, Options) of
        {ok, Ref} ->
            io:format("OK ==> ~p~n", [Ref]);
        {error, Reason} ->
            io:format("Error: ~p~n", [Reason])
    end,
    {continue, State};
eval_line({stop, Proto, From, To}, _RelayMgr, State) ->
    trelay_mgr:stop_relay(Proto, From, To),
    io:format("OK. Stopped.~n"),
    {continue, State};
eval_line({stop, Ref}, _RelayMgr, State) ->
    trelay_mgr:stop_relay(Ref),
    io:format("OK. Stopped.~n"),
    {continue, State};
eval_line({stop, RelayRef, ConnRef}, _RelayMgr, State) ->
    trelay_mgr:stop_relay({RelayRef, ConnRef}),
    io:format("OK. Stopped.~n"),
    {continue, State};
eval_line(stopall, _RelayMgr, State) ->
    R = trelay_mgr:stop_all_relays(),
    io:format("OK. All relays stopped: ~p~n", [R]),
    {continue, State};
eval_line({change_options, RelayRef, NewOptions}, _RelayMgr, State) ->
    R = trelay_mgr:change_options(RelayRef, NewOptions),
    io:format("OK. ~p.~n", [R]),
    {continue, State};
eval_line(list, _RelayMgr, State) ->
    case trelay_mgr:running_relays() of
        [] ->
            io:format("There are no running relays.~n");
        RelayList ->
            io:format("Relays:~n"),
            pretty_print_relays(RelayList)
    end,
    {continue, State};
eval_line({save, FileName}, _RelayMgr, State) ->
    R = trelay_mgr:save(FileName),
    io:format("OK: ~p.~n", [R]),
    {continue, State};
eval_line({load, FileName}, _RelayMgr, State) ->
    R = trelay_mgr:load(FileName),
    io:format("OK: ~p.~n", [R]),
    {continue, State};
eval_line({erlang, {Module, Function, Args}}, _RelayMgr, State) ->
    R = (catch apply(Module, Function, Args)),
    io:format("==>~p~n", [R]),
    {continue, State};
eval_line(dont_halt_on_eof, _RelayMgr, State) ->
    io:format("Not halting on eof.~n."),
    {continue, State#state{halt_on_eof = false}};
eval_line(help, _RelayMgr, State) ->
    print_help(),
    {continue, State};
eval_line(Other, _RelayMgr, State) ->
    io:format("Unexpected command: ~p~n", [Other]),
    {continue, State}.


pretty_print_relays(RelayList) ->
    lists:foreach(fun(Relay) -> pretty_print_relay(Relay) end, RelayList).

pretty_print_relay({Ref, Proto, From, To, Connections, Options}) ->
    io:format("  ~-4w ~p-", [Ref, Proto]),
    pretty_print_addr(From),
    io:format("-"),
    pretty_print_addr(To),
    lists:foreach(fun(Option) ->
                          io:format(" ~p", [Option])
                  end, Options),
    io:format("~n"),
    pretty_print_connections(Connections);
pretty_print_relay(OtherRelayInfo) ->
    io:format("  ~p~n", [OtherRelayInfo]).


pretty_print_connections([]) ->
    io:format("       No open connections~n");
pretty_print_connections(Connections) ->
    io:format("       Connections:~n"),
    lists:foreach(fun(Connection) -> pretty_print_connection(Connection) end,
                  Connections).

pretty_print_connection({RAddr, RPort, Since, Ref}) ->
    io:format("         ~-4w from ", [Ref]),
    pretty_print_addr({RAddr, RPort}),
    io:format(" opened "),
    pretty_print_time(Since),
    io:format("~n").



pretty_print_addr({any, PortNumber}) ->
    io:format("*:~p", [PortNumber]);
pretty_print_addr({{IP1,IP2,IP3,IP4}, PortNumber}) ->
    io:format("~p.~p.~p.~p:~p", [IP1,IP2,IP3,IP4, PortNumber]);
pretty_print_addr({HostName, PortNumber}) ->
    io:format("~s:~p", [HostName, PortNumber]).

pretty_print_time({{_YYYY,_MM,_DD}=Date, {_H,_M,_S}=Time}) ->
    pretty_print_time({Date, Time}, true).

%pretty_print_datetime({{YYYY,MM,DD}=Date, {H,M,S}=Time}) ->
%    pretty_print_time({Date, Time}, false).

pretty_print_time({{YYYY,MM,DD}=Date, {H,M,S}=_Time}, SuppressDate) ->
    case {SuppressDate, date()} of
        {true, Date} ->
            % Same date. Only print time
            io:format("~2w:~2..0w:~2..0w", [H,M,S]);
        _Otherwise ->
            io:format("~w-~2..0w-~2..0w ~2w:~2..0w:~2..0w", [YYYY,MM,DD,H,M,S])
    end.


print_welcome_msg() ->
    io:format(
"trelay version ~s, Copyright 2010 Tomas Abrahamsson <tab@lysator.liu.se>~n"
"trelay comes with ABSOLUTELY NO WARRANTY; for details type `show warranty'.~n"
"This is free software, and you are welcome to redistribute it~n"
"under certain conditions; type `show copying' for details.~n"
"~n"
"For help, type `help'.~n"
"~n",
      [version:version()]).


print_help() ->
    io:format(
"The following commands are available:~n"
"~n"
"  start Proto-FromAddr-ToAddr [Option ...]~n"
"    Start a relay for protocol~n"
"    Types:~n"
"      Proto = tcp | udp~n"
"      FromAddr = ToAddr = Host:Port~n"
"      Host = * | hostname() | ipnumber()~n"
"      Option = Expr[,Expr...]~n"
"~n"
"    The following options are available:~n"
"      {bwlimit,Limit}~n"
"        Limits the bandwidth by pulsing data at a specific rate.~n"
"        It is only possible to limit the bandwidth for tcp relays.~n"
"        This option is ignored for udp relays.~n"
"        Default is no bandwidth limiting.~n"
"        Types:~n"
"          Limit={Number, Unit}~n"
"          Unit=\"bps\" | \"baud\" | \"kbps\" | \"Mbps\"~n"
"      {quantum_length,Length}~n"
"        Set the length of the time quanta when pulsing data.~n"
"        Shorter length means smoother data flow, which uses~n"
"        more CPU-power, and vice versa.~n"
"        Default is a quantum length of 50 ms.~n"
"        Types:~n"
"          Length=integer() -- milliseconds per quantum~n"
"      dump~n"
"        Dump communications in hex/ascii form.~n"
"    Examples:~n"
"      start tcp-*:2222-localhost:22~n"
"        -- start a relay from port 2222 to port 22~n"
"      start tcp-localhost:2223-remotehost.remotedomain:22~n"
"        -- start a relay which listens on port 2223~n"
"           on the ip-address for localhost. Connections~n"
"           are relayed to remotehost.remotedomain, port 22~n"
"      start tcp-*:2224-localhost:22 dump~n"
"        -- start a relay from port 2224 to port 22 and~n"
"           dump all communication on the connection~n"
"      start tcp-*:2224-localhost:22 {bwlimit,{56700,\"bps\"}}~n"
"        -- start a relay from port 2224 to port 22 with~n"
"           bandwidth limited to 56700 kbit/s (modem speed).~n"
"~n"
"~n"
"  stop RelayReference~n"
"    Stops a relay (and any connections for that relay)~n"
"    by giving the relay's reference.~n"
"    Types:~n"
"      RelayReference = integer()~n"
"~n"
"  stop RelayReference ConnectionReference~n"
"    Stops a connection for a relay by giving the references~n"
"    for the relay and connection~n"
"    Types:~n"
"      RelayReference = ConnectionReference = integer()~n"
"~n"
"  changeopts RelayReference [Option ...]~n"
"    Change the options for a relay (and any connections for that relay)~n"
"    Types:~n"
"      RelayReference = integer()~n"
"      Options = same as for the start command~n"
"~n"
"  stopall~n"
"    Stops all started relays.~n"
"~n"
"  list~n"
"    Lists all running relays.~n"
"~n"
"  save FileName~n"
"    Save the relay setup to a file.~n"
"~n"
"  load FileName~n"
"    Load the relay setup from a file and start the relays.~n"
"~n"
"  halt~n"
"  quit~n"
"    Stops all running relays, then exits.~n"
"~n"
"  help~n"
"  ?~n"
"    Prints this message.~n"
"~n"
"  # comment-text~n"
"    Lines starting with the character `#' are comments.~n"
"~n").




input_start() ->
    spawn_link(?MODULE, input_init, [self()]).


input_init(CmdLineServer) ->
    receive
        start ->
            input_loop(CmdLineServer)
    end.


input_loop(CmdLineServer) ->
    case io:get_line('> ') of
        eof ->
            CmdLineServer ! {line, self(), eof},
            input_loop_eof(CmdLineServer);
        Line ->
            LineNoNL = case lists:reverse(Line) of
                           [$\n | LineReversed] ->
                               lists:reverse(LineReversed);
                           _NoTrailingNL ->
                               Line
                       end,
            CmdLineServer ! {line, self(), LineNoNL},
            receive
                continue ->
                    input_loop(CmdLineServer);
                stop ->
                    done
            end
    end.

input_loop_eof(CmdLineServer) ->
    receive
        continue ->
            CmdLineServer ! {line, self(), eof},
            input_loop_eof(CmdLineServer);
        stop ->
            done
    end.


%% ----------------------------------------------------------------------
%% list_to_term -- string() to term()
%% Returns:     {ok, Term} | {error, ErrorInfo}
%% ----------------------------------------------------------------------
list_to_term(StringNoDot) ->
    case erl_scan:string(StringNoDot ++ ".") of
        {ok, Tokens, _EndLine} ->
            %% This returns {ok, Term} | {error, ErrorInfo}
            erl_parse:parse_term(Tokens);
        {error, ErrorInfo, _EndLine} ->
            {error, ErrorInfo}
    end.
