%%%----------------------------------------------------------------------
%%% File    : trelay.erl
%%% Author  : Tomas Abrahamsson <tab@lysator.liu.se>
%%% Purpose : TCP/UDP relay program
%%% Created : 10 May 1980 by Tomas Abrahamsson <tab@lysator.liu.se>
%%%----------------------------------------------------------------------

%%%----------------------------------------------------------------------
%%% trelay -- a program for relaying tcp/udp traffic
%%% Copyright (C) 2000  Tomas Abrahamsson <tab@lysator.liu.se>
%%%
%%% This program is free software; you can redistribute it and/or modify
%%% it under the terms of the GNU General Public License as published by
%%% the Free Software Foundation; either version 2 of the License, or
%%% (at your option) any later version.
%%%
%%% This program is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU General Public License for more details.
%%%
%%% You should have received a copy of the GNU General Public License
%%% along with this program; if not, write to the Free Software
%%% Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
%%%----------------------------------------------------------------------

-module(trelay).
-author('tab@lysator.liu.se').
-vsn('$Id$').   % '

-export([main/0]).
-export([main/1]).

%%-compile(export_all).
%%-export([Function/Arity, ...]).
-export([start/3, start/4, start_link/3, start_link/4]).
-export([change_options/2, stop_connection/1]).

%% exports due to spawns
-export([tcp_server_init/4, tcp_link_init/5]).
-export([udp_init/4]).

main() ->
    main([]).
main(_) ->
    main:main().


-record(opts, {is_master = true,                % true | false
               announce_to = none,              % {pid(), Ref}
               connection_ref = none,           % #Ref{}
               l_sock_addr,
               l_sock_addr_text,
               l_peer_addr,
               l_peer_addr_text,
               r_sock_addr,
               r_sock_addr_text,
               r_peer_addr,
               r_peer_addr_text,
               bw = unlimited,
               dump = false,
               quantum_length = 50,             % millisecs
               max_bytes_per_quantum = 100}).

%% ------------------------------------------------------------------
%% start, start_link -- start a relay
%% Arguments:   Proto = udp | tcp
%%              From = {FromAddress, PortNumber}
%%                FromAddress = any | IPNumber | HostString
%%                IPNumber = {int(), int(), int(), int()}
%%                HostString = string()
%%              To = {Address, PortNumber}
%%                Address = IPNumber | string()
%% Returns:     {ok, RelayPid} | {error, Reason}
%% ------------------------------------------------------------------
start(Proto, From, To) ->
    start(Proto, From, To, []).

start(udp, From, To, Options) ->
    Name = compute_name(udp, From, To),
    RelayServer = spawn(?MODULE, udp_init, [self(), From, To, Options]),
    register_started(Name, RelayServer);
start(tcp, From, To, Options) ->
    Name = compute_name(tcp, From, To),
    RelayServer = spawn(?MODULE, tcp_server_init, [self(), From, To, Options]),
    register_started(Name, RelayServer).


start_link(Proto, From, To) ->
    start_link(Proto, From, To, []).

start_link(udp, From, To, Options) ->
    Name = compute_name(udp, From, To),
    RelayServer = spawn_link(?MODULE, udp_init, [self(), From, To, Options]),
    register_started(Name, RelayServer);
start_link(tcp, From, To, Options) ->
    Name = compute_name(tcp, From, To),
    RelayServer = spawn_link(?MODULE,
                             tcp_server_init, [self(), From, To, Options]),
    register_started(Name, RelayServer).


change_options(RelayOrConnPid, NewOptions) ->
    NewOpts = parse_options(NewOptions),
    RelayOrConnPid ! {change_opts, NewOpts}.

stop_connection(ConnMasterPid) ->
    ConnMasterPid ! ordered_stop.


register_started(Name, RelayServer) ->
    register(Name, RelayServer),
    receive
        ok ->
            {ok, RelayServer};
        {error, Reason} ->
            {error, Reason}
    end.


%% ----------------------------------------------------------------------
%% ----------------------------------------------------------------------
%% TCP code
%% ----------------------------------------------------------------------
%% ----------------------------------------------------------------------

tcp_server_init(Master, {FromAddr, PortNumber}, To, Options) ->
    Opts = parse_options(Options),
    AddrOpt = addr_to_listen_opt(FromAddr),
    ListenOpts = [binary,
                  {packet, 0},
                  {active, false},
                  {nodelay, true},
                  {reuseaddr, true}] ++ AddrOpt,
    case gen_tcp:listen(PortNumber, ListenOpts) of
        {ok, ServerSocket} ->
            Master ! ok,
            tcp_server_loop(Master, To, ServerSocket, Opts);
        {error, Reason} ->
            Master ! {error, {listen_failed, Reason}}
    end.

tcp_server_loop(Master, ToAddr, ServerSocket, Opts) ->
    tcp_server_loop(Master, ToAddr, ServerSocket, Opts, 0).
tcp_server_loop(Master, ToAddr, ServerSocket, Opts, NextRef) ->
    case gen_tcp:accept(ServerSocket, 250) of
        {ok, ClientSocket} ->
            NewOpts = Opts#opts{connection_ref = NextRef},
            tcp_link_start(Master, ClientSocket, ToAddr, NewOpts),
            tcp_server_loop(Master, ToAddr, ServerSocket, Opts, NextRef + 1);
        {error, timeout} ->
            tcp_server_loop_check(Master, ToAddr, ServerSocket, Opts, NextRef);
        {error, Reason} ->
            log("Error: accepting on ~p: ~p.", [ServerSocket, Reason])
    end.

tcp_server_loop_check(Master, ToAddr, ServerSocket, Opts, NextRef) ->
    receive
        {change_opts, NewOpts} ->
            tcp_server_loop(Master, ToAddr, ServerSocket, NewOpts, NextRef)
    after 0 ->
            tcp_server_loop(Master, ToAddr, ServerSocket, Opts, NextRef)
    end.


%% Wrapper around announce/2
announce_new_connection(ClientSocket, Opts) ->
    case inet:peername(ClientSocket) of
        {ok, {_RemoteAddress, _RemotePort}=RemoteSide} ->
            ConnRef = Opts#opts.connection_ref,
            Announcement = {new_connection, self(), ConnRef, RemoteSide},
            announce(Opts, Announcement);
        {error, _Reason} ->
            ignore_the_error
    end.

%% Wrapper around announce/2
announce_closed_connection(Opts) ->
    ConnRef = Opts#opts.connection_ref,
    Announcement = {closed_connection, ConnRef},
    announce(Opts, Announcement).


%% Announce
announce(Opts, Announcement) ->
    case {Opts#opts.is_master, Opts#opts.announce_to} of
        {true, {Pid, UserKey}} ->
            Pid ! {trelay_announcement, UserKey, Announcement};
        _ ->
            dont_annonunce
    end.


tcp_link_start(Master, LSocket, {RHostName, RPortNum} = RemoteAddr, Opts) ->
    RHost = compute_rhost(RHostName, LSocket),
    RSockOpts = [binary, {packet, 0}, {nodelay, true}, {active, false}],
    case gen_tcp:connect(RHost, RPortNum, RSockOpts) of
        {ok, RSocket} ->
            {ok, LSockAddr} = inet:sockname(LSocket),
            {ok, LPeerAddr} = inet:peername(LSocket),
            {ok, RSockAddr} = inet:sockname(RSocket),
            {ok, RPeerAddr} = inet:peername(RSocket),
            LSockAddrTxt = compute_name_addr(LSockAddr),
            LPeerAddrTxt = compute_name_addr(LPeerAddr),
            RSockAddrTxt = compute_name_addr(RSockAddr),
            RPeerAddrTxt = compute_name_addr(RPeerAddr),
            SockPeerOpts = Opts#opts{l_sock_addr      = LSockAddr,
                                     l_sock_addr_text = LSockAddrTxt,
                                     l_peer_addr      = LPeerAddr,
                                     l_peer_addr_text = LPeerAddrTxt,
                                     r_sock_addr      = RSockAddr,
                                     r_sock_addr_text = RSockAddrTxt,
                                     r_peer_addr      = RPeerAddr,
                                     r_peer_addr_text = RPeerAddrTxt},
            LToROpts = SockPeerOpts#opts{is_master = true},
            RToLOpts = SockPeerOpts#opts{is_master = false},
            maybe_dump(new_connection, LToROpts, LSocket, RSocket, RemoteAddr),
            LtoRArgs = [Master, self(), LSocket, RSocket, LToROpts],
            RtoLArgs = [Master, self(), RSocket, LSocket, RToLOpts],
            LtoRHandler = spawn_link(?MODULE, tcp_link_init, LtoRArgs),
            RtoLHandler = spawn_link(?MODULE, tcp_link_init, RtoLArgs),
            LtoRHandler ! {link_handler_other_direction, RtoLHandler},
            RtoLHandler ! {link_handler_other_direction, LtoRHandler},
            gen_tcp:controlling_process(LSocket, LtoRHandler),
            gen_tcp:controlling_process(RSocket, RtoLHandler),
            LtoRHandler ! go_active,
            RtoLHandler ! go_active,
            {LtoRHandler, RtoLHandler};

        {error, Reason} ->
            log("Error opening socket to remote address: ~p.~n", [Reason]),
            gen_tcp:close(LSocket)
    end.


compute_rhost(same, LocalSocket) ->
    {ok, {LocalIP, _Port}} = inet:sockname(LocalSocket),
    LocalIP;
compute_rhost(RHostName, _LSocket) ->
    RHostName.


tcp_link_init(_Master, _Server, LocalSocket, RemoteSocket, Opts) ->
    %% Do some book-keeping
    OtherHandler =
        receive
            {link_handler_other_direction, OtherLinkHandler} ->
                announce_new_connection(LocalSocket, Opts),
                OtherLinkHandler
        end,
    %% Handle incoming data on the socket as messages.
    receive
        go_active ->
            inet:setopts(LocalSocket, [{active, true}]),
            tcp_link_loop(OtherHandler, LocalSocket, RemoteSocket, Opts)
    end.





% Forwards data from local-socket to remote-socket.
tcp_link_loop(OtherHandler, LocalSocket, RemoteSocket, Opts) ->
    receive
        {tcp, LocalSocket, Data} ->
            maybe_dump(got_data, Opts, LocalSocket, RemoteSocket, Data),
            tcp_link_send(OtherHandler, LocalSocket, RemoteSocket, Data, Opts);
        {tcp_closed, LocalSocket} ->
            maybe_dump(got_close, Opts, LocalSocket, RemoteSocket),
            OtherHandler ! do_close_socket,
            announce_closed_connection(Opts);
        do_close_socket ->
            %% Other side orders close down
            gen_tcp:close(LocalSocket);
        ordered_stop ->
            %% Only the master gets this message.
            tcp_link_stop(OtherHandler, LocalSocket, Opts);
        {change_opts, NewOpts} ->
            %% Only the master gets this message.
            MyNewOpts = NewOpts#opts{is_master = Opts#opts.is_master},
            OtherOpts = NewOpts#opts{is_master = not Opts#opts.is_master},
            OtherHandler ! {do_change_opts, OtherOpts},
            tcp_link_loop(OtherHandler, LocalSocket, RemoteSocket, MyNewOpts);
        {do_change_opts, NewOpts} ->
            %% Other side (=master) orders opts change
            tcp_link_loop(OtherHandler, LocalSocket, RemoteSocket, NewOpts);
        {tcp_error, LocalSocket, Reason} ->
            log("Socket error on ~p (client): ~p.~n", [LocalSocket, Reason]),
            gen_tcp:close(LocalSocket),
            OtherHandler ! link_tcp_error;
        link_tcp_error ->
            gen_tcp:close(LocalSocket);
        Other ->
            log("Other event?!: ~p~n", [Other]),
            tcp_link_loop(OtherHandler, LocalSocket, RemoteSocket, Opts)
    end.

tcp_link_send(OtherHandler, LSocket, RSocket, Data, Opts) ->
    case Opts#opts.bw of
        unlimited ->
            gen_tcp:send(RSocket, Data),
            tcp_link_loop(OtherHandler, LSocket, RSocket, Opts);
        limited ->
            tcp_link_send_bwlimit(OtherHandler, LSocket, RSocket, Data, Opts)
    end.

tcp_link_send_bwlimit(OtherHandler, LSocket, RSocket, Data, Opts) ->
    %% Assumption1: The time spent to call gen_tcp:send is negligable
    %% Assumption2: gen_tcp:send does not block until all is sent.
    %%              This seems to be true from the source code for the
    %%              tcp/udp driver.
    QMax = Opts#opts.max_bytes_per_quantum,
    QLen = Opts#opts.quantum_length,
    if
        size(Data) > QMax ->
            {ToSendNow, Later} = split_binary(Data, QMax),
            gen_tcp:send(RSocket, ToSendNow),
            sleep(QLen),
            tcp_link_send_check(OtherHandler, LSocket, RSocket, Later, Opts);
        true ->
            gen_tcp:send(RSocket, Data),
            ShortenedQLen = round(size(Data) / QMax * QLen),
            sleep(ShortenedQLen),
            tcp_link_loop(OtherHandler, LSocket, RSocket, Opts)
    end.

sleep(MilliSecs) ->
    receive after MilliSecs -> ok end.

tcp_link_send_check(OtherHandler, LSocket, RSocket, Data, Opts) ->
    receive
        ordered_stop ->
            %% Only the master gets this message.
            tcp_link_stop(OtherHandler, LSocket, Opts);
        {change_opts, NewOpts} ->
            %% Only the master gets this message.
            MyNewOpts = NewOpts#opts{is_master = Opts#opts.is_master},
            OtherOpts = NewOpts#opts{is_master = not Opts#opts.is_master},
            OtherHandler ! {do_change_opts, OtherOpts},
            tcp_link_send(OtherHandler, LSocket, RSocket, Data, MyNewOpts);
        {do_change_opts, NewOpts} ->
            %% Other side (=master) orders opts change
            tcp_link_send(OtherHandler, LSocket, RSocket, Data, NewOpts)
    after 0 ->
            tcp_link_send(OtherHandler, LSocket, RSocket, Data, Opts)
    end.


tcp_link_stop(OtherHandler, LocalSocket, Opts) ->
    gen_tcp:close(LocalSocket),
    OtherHandler ! do_close_socket,
    announce_closed_connection(Opts).


%% ----------------------------------------------------------------------
%% ----------------------------------------------------------------------
%% UDP code
%% ----------------------------------------------------------------------
%% ----------------------------------------------------------------------
udp_init(Master, {FromAddr, FromPort}=_From, {ToAddr, ToPort}=_To, _Options) ->
    FromAddrOpt  = addr_to_listen_opt(FromAddr),
    FromSockOpts = [binary, {packet, 0}] ++ FromAddrOpt,
    ToSockOpts   = [binary, {packet, 0}],
    FromResult   = gen_udp:open(FromPort, FromSockOpts),
    ToResult     = gen_udp:open(0, ToSockOpts),
    case {FromResult, ToResult} of
        {{error, Reason1}, {error, Reason2}} ->
            Master ! {error, {Reason1, Reason2}};
        {{error, Reason1}, {ok, ToSocket}} ->
            Master ! {error, {Reason1}},
            gen_udp:close(ToSocket);
        {{ok, FromSocket}, {error, Reason2}} ->
            Master ! {error, {Reason2}},
            gen_udp:close(FromSocket);
        {{ok, FromSocket}, {ok, ToSocket}} ->
            Master ! ok,
            udp_loop(Master,
                     {FromAddr, FromPort, FromSocket},
                     {ToAddr, ToPort, ToSocket})
    end.

udp_loop(Master,
         {FromAddr, FromPort, FromSocket} = From,
         {ToAddr, ToPort, ToSocket} = To) ->
    receive
        {udp, FromSocket, _OtherIP, _OtherPortNum, Packet} ->
            gen_udp:send(ToSocket, ToAddr, ToPort, Packet),
            udp_loop(Master, From, To);
        {udp, ToSocket, _OtherIP, _OtherPortNum, Packet} ->
            %% FIXME: FromSocket??
            gen_udp:send(ToSocket, FromAddr, FromPort, Packet),
            udp_loop(Master, From, To)
    end.


%% ----------------------------------------------------------------------
%% ----------------------------------------------------------------------
%% Common internal routines
%% ----------------------------------------------------------------------
%% ----------------------------------------------------------------------

addr_to_listen_opt(Addr) ->
    case Addr of
        any ->
            [];
        {_IP1, _IP2, _IP3, _IP3} = FromIP ->
            {ip, FromIP};
        HostString ->
            case inet:getaddr(HostString, inet) of
                {ok, IPAddr} ->
                    [{ip, IPAddr}];
                {error, Reason} ->
                    log("Error retrieving ip for ~p: ~p~n", [Addr, Reason]),
                    []
            end
    end.


compute_name(Proto, From, To) ->
    list_to_atom("trelay-" ++
                 atom_to_list(Proto) ++ "-" ++
                 compute_name_addr(From) ++ "-" ++
                 compute_name_addr(To)).


compute_name_addr({any, Port}) ->
    "*:" ++ integer_to_list(Port);
compute_name_addr({same, Port}) ->
    "=:" ++ integer_to_list(Port);
compute_name_addr({{IP1, IP2, IP3, IP4}, Port}) ->
    integer_to_list(IP1) ++ "." ++
        integer_to_list(IP2) ++ "." ++
        integer_to_list(IP3) ++ "." ++
        integer_to_list(IP4) ++ ":" ++
        integer_to_list(Port);
compute_name_addr({HostStr, Port}) ->
    HostStr ++ ":" ++ integer_to_list(Port).


parse_options(Options) ->
    Opts = #opts{},
    parse_options(Options, Opts).

parse_options([], Opts) ->
    Opts;
parse_options([{quantum_length, NewQLenMS} | RestOfOptions], Opts) ->
    CurrQLenMS = Opts#opts.quantum_length,
    CurrQLenS = CurrQLenMS / 1000,
    CurrQMax = Opts#opts.max_bytes_per_quantum,
    NewQLenS = NewQLenMS / 1000,
    CurrBWLimit = CurrQLenS * CurrQMax,
    NewQMax = round(CurrBWLimit * NewQLenS),
    NewOpts = Opts#opts{quantum_length = NewQLenMS,
                        max_bytes_per_quantum = NewQMax},
    parse_options(RestOfOptions, NewOpts);
parse_options([{bwlimit, BWLimit} | RestOfOptions], Opts) ->
    NewBWLimit = parse_bw_limit(BWLimit),
    CurrQLenMS = Opts#opts.quantum_length,
    CurrQLenS = CurrQLenMS / 1000,
    NewQMax = round(NewBWLimit * CurrQLenS),
    NewOpts = Opts#opts{bw = limited,
                        max_bytes_per_quantum = NewQMax},
    parse_options(RestOfOptions, NewOpts);
parse_options([{announce_to, {Pid, RelayRef}} | RestOfOptions], Opts) ->
    NewOpts = Opts#opts{announce_to = {Pid, RelayRef}},
    parse_options(RestOfOptions, NewOpts);
parse_options([dump | RestOfOptions], Opts) ->
    NewOpts = Opts#opts{dump = true},
    parse_options(RestOfOptions, NewOpts);
parse_options([OtherOption | RestOfOptions], Opts) ->
    log("Ignored option: ~p~n", [OtherOption]),
    parse_options(RestOfOptions, Opts).


parse_bw_limit({N, "bps"}) ->                   % bits per second
    round(N / 8);
parse_bw_limit({N, "baud"}) ->                  % alias for bps
    round(N / 8);
parse_bw_limit({N, "kbps"}) ->                  % kilobits per second
    round(N * 1000 / 8);
parse_bw_limit({N, "Mbps"}) ->                  % Megabits per second
    round(N * 1000000 / 8);
parse_bw_limit({_N, _OtherUnit}) ->
    {error, unknown_unit}.


maybe_dump(new_connection, Opts, _LSock, _RSock, _RAddr)
  when Opts#opts.dump == true ->
    LocalSockAddrTxt = Opts#opts.l_sock_addr_text,
    LocalPeerAddrTxt = Opts#opts.l_peer_addr_text,
    ToAddrTxt = Opts#opts.r_peer_addr_text,
    log("~nNew connection: ~s (from ~s) --> ~s~n",
        [LocalSockAddrTxt, LocalPeerAddrTxt, ToAddrTxt]);
maybe_dump(new_connection, _Opts, _LSock, _RSock, _RAddr) ->
    do_nothing;


maybe_dump(got_data, Opts, _LSock, _RSock, Data) when Opts#opts.dump == true ->
    LSockTxt = Opts#opts.l_sock_addr_text,
    LPeerTxt = Opts#opts.l_peer_addr_text,
    RPeerTxt = Opts#opts.r_peer_addr_text,
    Heading = case Opts#opts.is_master of
                  true  -> fmt("--> ~-20.. s (from ~-20.. s) >> ~s",
                               [LSockTxt, LPeerTxt, RPeerTxt]);
                  false -> fmt("<-- ~-20.. s (from ~-20.. s) << ~s",
                               [LSockTxt, LPeerTxt, RPeerTxt])
                end,
    HexData = hex_dump(Data),
    log("~n~s~n~s~n", [Heading, HexData]);
maybe_dump(got_data, _Opts, _LSocket, _RSocket, _Data) ->
    do_nothing.


maybe_dump(got_close, Opts, _LSock, _RSock) when Opts#opts.dump == true ->
    Direction = case Opts#opts.is_master of
                    true  -> "-->";
                    false -> "<--"
                end,
    LSockTxt = Opts#opts.l_sock_addr_text,
    LPeerTxt = Opts#opts.l_peer_addr_text,
    log("~n~s ~s other end (~s) closed the connection.~n",
        [Direction, LSockTxt, LPeerTxt]);
maybe_dump(got_close, _Opts, _LSock, _RSock) ->
    do_nothing.


%% ----------------------------------------------------------------------
%% log -- log a text (currnetly to stdout)
%% ----------------------------------------------------------------------
fmt(Fmt, Args) ->
    lists:flatten(io_lib:format(Fmt, Args)).

%log(Text) ->
%    io:format(Text).
log(Fmt, Data) ->
    io:format(Fmt, Data).


hex_dump(Data) when is_binary(Data) ->
    hex_dump(binary_to_list(Data));
hex_dump(Data) when length(Data) rem 16 /= 0 ->
    PadLen = 16 - (length(Data) rem 16),
    Pad = hex_dump_do_pad(PadLen),
    hex_dump(Data ++ Pad);
hex_dump(Data) ->
    lists:flatten(hex_dump(0, lists:flatten(Data), "")).


hex_dump(_, [], ReversedLines) ->
    lists:reverse(ReversedLines);
hex_dump(Adr, [D0,D1,D2,D3,D4,D5,D6,D7,D8,D9,DA,DB,DC,DD,DE,DF | Rest], Txt) ->
    H = int2hex(Adr),
    S0 = io_lib:format("  ~4.*.0s:", [length(H), H]),
    S1 = hex_dump_digits([D0,D1,D2,D3,D4,D5,D6,D7,D8,D9,DA,DB,DC,DD,DE,DF]),
    S2 = "   ",
    S3 = hex_dump_chars([D0,D1,D2,D3,D4,D5,D6,D7,D8,D9,DA,DB,DC,DD,DE,DF]),
    S4 = io_lib:format("~n", []),
    Line = S0 ++ S1 ++ S2 ++ S3 ++ S4,
    hex_dump(Adr + 16, Rest, [Line | Txt]).


hex_dump_digits(Digits) ->
    lists:map(fun(none) -> "   ";
                 (D) -> H = int2hex(D),
                        io_lib:format(" ~2.*.0s", [length(H), H])
              end,
              Digits).


hex_dump_chars(Digits) ->
    lists:map(fun(none) -> "";
                 (C) when C < 32 -> ".";
                 (C) when C >= 128, C < 160 -> ".";
                 (C) -> [C]
              end,
              Digits).

hex_dump_do_pad(PadLen) ->
    lists:duplicate(PadLen, none).


int2hex(N) when N < 0 ->
        "-" ++ int2hex(-N);
int2hex(N) ->
        int2hex(N, []).

int2hex(0, Acc) ->
        Acc;
int2hex(N, Acc) ->
        int2hex(N div 16, [i2h(N rem 16)|Acc]).

i2h(N) when N < 10 ->
        $0 + N;
i2h(N) ->
        $a + N - 10.

