%%%----------------------------------------------------------------------
%%% File    : trelay_mgr.erl
%%% Author  : Tomas Abrahamsson <tab@lysator.liu.se>
%%% Purpose :
%%% Created : 25 May 2000 by Tomas Abrahamsson <tab@lysator.liu.se>
%%%----------------------------------------------------------------------

%%%----------------------------------------------------------------------
%%% trelay -- a program for relaying tcp/udp traffic
%%% Copyright (C) 2000  Tomas Abrahamsson <tab@lysator.liu.se>
%%%
%%% This program is free software; you can redistribute it and/or modify
%%% it under the terms of the GNU General Public License as published by
%%% the Free Software Foundation; either version 2 of the License, or
%%% (at your option) any later version.
%%%
%%% This program is distributed in the hope that it will be useful,
%%% but WITHOUT ANY WARRANTY; without even the implied warranty of
%%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%%% GNU General Public License for more details.
%%%
%%% You should have received a copy of the GNU General Public License
%%% along with this program; if not, write to the Free Software
%%% Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
%%%----------------------------------------------------------------------

-module(trelay_mgr).
-vsn('$Id$').   % '
-author('tab@lysator.liu.se').

%%-compile(export_all).
%%-export([Function/Arity, ...]).

-behaviour(gen_server).

%% External exports
-export([start/0, start_link/0, stop/0]).
-export([start_relay/4, stop_relay/1, stop_relay/3, stop_all_relays/0]).
-export([change_options/4, change_options/2]).
-export([running_relays/0]).
-export([save/1, load/1]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2]).
-export([code_change/3]).

%%%----------------------------------------------------------------------
%%% API
%%%----------------------------------------------------------------------
start() ->
    gen_server:start({local, ?MODULE}, ?MODULE, [], []).
start_link() ->
    gen_server:start_link({local, ?MODULE}, ?MODULE, [], []).

stop() ->
    gen_server:call(?MODULE, stop).


%% Returns:     RelayReferece
start_relay(Proto, From, To, Options) ->
    gen_server:call(?MODULE, {start_relay, Proto, From, To, Options}).

stop_relay({RelayRef, ConnRef}) ->
    gen_server:call(?MODULE, {stop_relay, {RelayRef, ConnRef}});
stop_relay(RelayRef) ->
    gen_server:call(?MODULE, {stop_relay, RelayRef}).
stop_relay(Proto, From, To) ->
    gen_server:call(?MODULE, {stop_relay, Proto, From, To}).

stop_all_relays() ->
    gen_server:call(?MODULE, stop_all_relays).

change_options(Proto, From, To, NewOptions) ->
    gen_server:call(?MODULE, {change_options, Proto, From, To, NewOptions}).
change_options(RelayRef, NewOptions) ->
    gen_server:call(?MODULE, {change_options, RelayRef, NewOptions}).

running_relays() ->
    gen_server:call(?MODULE, running_relays).

save(FileName) ->
    gen_server:call(?MODULE, {save, FileName}).

load(FileName) ->
    gen_server:call(?MODULE, {load, FileName}).

%%%----------------------------------------------------------------------
%%% Callback functions from gen_server
%%%----------------------------------------------------------------------


-record(state, {table_id = none, next_ref = 0}).
-record(relay, {ref = none,             % int()
                proto = none,
                from = none,
                to = none,
                options = none,
                pid = none,             % pid()
                connections = []}).
-record(connection, {r_addr = none,
                     r_port = none,
                     since = none,              % when it was started
                     pid = none,
                     ref = none}).

% Relay as saved to file
-record(file_relay, {proto = none,
                     from = none,
                     to = none,
                     options = none}).


%%----------------------------------------------------------------------
%% Func: init/1
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          ignore               |
%%          {stop, Reason}
%%----------------------------------------------------------------------
init([]) ->
    process_flag(trap_exit, true),
    TableId = ets:new(?MODULE, [set, protected, {keypos, #relay.ref}]),
    {ok, #state{table_id = TableId}}.

%%----------------------------------------------------------------------
%% Func: handle_call/3
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_call({start_relay, Proto, From, To, Options}, _From, State) ->
    NewRef = State#state.next_ref,
    case start_relay(NewRef, Proto, From, To, Options) of
        {ok, NewRelay} ->
            ets:insert(State#state.table_id, NewRelay),
            {reply, {ok, NewRef}, State#state{next_ref = NewRef + 1}};
        {error, Reason} ->
            {reply, {error, Reason}, State}
    end;

handle_call({stop_relay, Proto, From, To}, _From, State) ->
    case find_relay(State#state.table_id, Proto, From, To) of
        {ok, Relay} ->
            do_stop_relay(Relay),
            ets:delete(State#state.table_id, Relay#relay.ref),
            {reply, stopped, State};
        _ ->
            {reply, relay_not_found, State}
    end;

handle_call({stop_relay, {RelayRef, ConnRef}}, _From, State) ->
    case find_relay(State#state.table_id, RelayRef) of
        {ok, Relay} ->
            stop_connection(Relay, ConnRef),
            {reply, stopped, State};
        _ ->
            {reply, relay_not_found, State}
    end;

handle_call({stop_relay, RelayRef}, _From, State) ->
    case find_relay(State#state.table_id, RelayRef) of
        {ok, Relay} ->
            do_stop_relay(Relay),
            ets:delete(State#state.table_id, Relay#relay.ref),
            {reply, stopped, State};
        _ ->
            {reply, relay_not_found, State}
    end;

handle_call(stop_all_relays, _From, State) ->
    lists:foreach(fun(Relay) ->
                          do_stop_relay(Relay),
                          ets:delete(State#state.table_id, Relay#relay.ref)
                  end,
                  ets:tab2list(State#state.table_id)),
    {reply, stopped, State};

handle_call({change_options, Proto, From, To, NewOptions}, _From, State) ->
    case find_relay(State#state.table_id, Proto, From, To) of
        {ok, Relay} ->
            NewRelay = do_change_options(Relay, NewOptions),
            ets:insert(State#state.table_id, NewRelay),
            {reply, ok, State};
        _ ->
            {reply, relay_not_found, State}
    end;

handle_call({change_options, RelayRef, NewOptions}, _From, State) ->
    case find_relay(State#state.table_id, RelayRef) of
        {ok, Relay} ->
            NewRelay = do_change_options(Relay, NewOptions),
            ets:insert(State#state.table_id, NewRelay),
            {reply, ok, State};
        _ ->
            {reply, relay_not_found, State}
    end;



handle_call(running_relays, _From, State) ->
    Relays = export_relays(State#state.table_id),
    {reply, Relays, State};

handle_call(stop, _From, State) ->
    lists:foreach(fun(Relay) ->
                          do_stop_relay(Relay),
                          ets:delete(State#state.table_id, Relay#relay.ref)
                  end,
                  ets:tab2list(State#state.table_id)),
    {stop, normal, ok, State};

handle_call({save, FileName}, _From, State) ->
    case file:open(FileName, [write]) of
        {ok, IoDevice} ->
            FileFormatVersion = {file_format_version, 1},
            DataToSave = to_file_relays(State#state.table_id),
            io:format(IoDevice, "~p.~n~p.~n", [FileFormatVersion, DataToSave]),
            file:close(IoDevice),
            {reply, ok, State};
        {error, Reason} ->
            {reply, {error, Reason}, State}
    end;

handle_call({load, FileName}, _From, State) ->
    case file:consult(FileName) of
        {ok, [VersionTerm, RelayTerms]} ->
            Relays = parse_relays(VersionTerm, RelayTerms),
            RunningRelays = ets:tab2list(State#state.table_id),
            NewRelays = extract_new_relays(Relays, RunningRelays),
            NewRef = State#state.next_ref,
            TableId = State#state.table_id,
            NewNextRef = start_relays(TableId, NewRef, NewRelays),
            {reply, ok, State#state{next_ref = NewNextRef}};
        {ok, _OtherTerms} ->
            {reply, data_error, State};
        {error, Reason} when is_atom(Reason) ->
            {reply, {file_error, Reason}, State};
        {error, Reason}  ->
            {reply, {parse_error, Reason}, State}
    end;

handle_call(Request, _From, State) ->
    Reply = {bad_request, Request},
    {reply, Reply, State}.

%%----------------------------------------------------------------------
%% Func: handle_cast/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_cast(_Msg, State) ->
    {noreply, State}.

%%----------------------------------------------------------------------
%% Func: handle_info/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_info({'EXIT', _RelayPid, stopped}, State) ->
    {noreply, State};

handle_info({'EXIT', RelayPid, Reason}, State) ->
    io:format("Relay ~p died: Reason:~n  ~p~n", [RelayPid, Reason]),
    TableId = State#state.table_id,
    case find_relay(TableId, RelayPid) of
        {ok, Relay} ->
            ets:delete(TableId, Relay#relay.ref);
        {error, _Reason} ->
            ignore
    end,
    {noreply, State};

handle_info({trelay_announcement, RelayRef, Announcement}, State) ->
    TableId = State#state.table_id,
    handle_announcement(RelayRef, Announcement, TableId),
    {noreply, State};

handle_info(Info, State) ->
    io:format("Info msg:  ~p~n", [Info]),
    {noreply, State}.

%%----------------------------------------------------------------------
%% Func: terminate/2
%% Purpose: Shutdown the server
%% Returns: any (ignored by gen_server)
%%----------------------------------------------------------------------
terminate(_Reason, _State) ->
    ok.

%%----------------------------------------------------------------------
%% Func: code_change/2
%% Purpose: Code update
%% Returns: {ok, NewState}
%%----------------------------------------------------------------------
code_change(_OldVsn, State, _Extra) ->
    {ok, State}.

%%%----------------------------------------------------------------------
%%% Internal functions
%%%----------------------------------------------------------------------

%%----------------------------------------------------------------------
%% Starting/stopping functions

start_relay(Ref, Proto, From, To, Options) ->
    StartOptions = add_announcement_option(Options, Ref),
    case trelay:start_link(Proto, From, To, StartOptions) of
        {ok, RelayPid} ->
            NewRelay = #relay{proto = Proto,
                              from = From,
                              to = To,
                              options = Options,
                              ref = Ref,
                              pid = RelayPid},
            {ok, NewRelay};
        {error, Reason} ->
            {error, Reason}
    end.


do_stop_relay(Relay) ->
    exit(Relay#relay.pid, stopped).


stop_connection(Relay, ConnRef) ->
    case find_connection(Relay#relay.connections, ConnRef) of
        {ok, Conn} ->
            trelay:stop_connection(Conn#connection.pid);
        _ ->
            ignore
    end.

%%----------------------------------------------------------------------
%% Finding relays and connections

find_relay(TableId, Proto, From, To) ->
    Results = [Relay || Relay <- ets:tab2list(TableId),
                        Relay#relay.proto == Proto,
                        Relay#relay.from == From,
                        Relay#relay.to == To],
    case Results of
        [Relay] ->
            {ok, Relay};
        [] ->
            {error, no_matches};
        _TooManyMatches ->
            {error, too_many_matches}
    end.

find_relay(TableId, RelayRef) when is_integer(RelayRef) ->
    Results = [Relay || Relay <- ets:tab2list(TableId),
                        Relay#relay.ref == RelayRef],
    case Results of
        [Relay] ->
            {ok, Relay};
        [] ->
            {error, no_matches};
        _TooManyMatches ->
            {error, too_many_matches}
    end;

find_relay(TableId, RelayPid) when is_pid(RelayPid) ->
    Results = [Relay || Relay <- ets:tab2list(TableId),
                        Relay#relay.pid == RelayPid],
    case Results of
        [Relay] ->
            {ok, Relay};
        [] ->
            {error, no_matches};
        _TooManyMatches ->
            {error, too_many_matches}
    end.

find_connection(Conns, ConnRef) ->
    Results = [Conn || Conn <- Conns,
                       Conn#connection.ref == ConnRef],
    case Results of
        [Conn] ->
            {ok, Conn};
        [] ->
            {error, no_matches};
        _TooManyMatches ->
            {error, too_many_matches}
    end.

sort_relays(Relays) ->
    lists:sort(fun(Relay1, Relay2) ->
                       Relay1#relay.ref < Relay2#relay.ref
               end,
               Relays).

sort_connections(Connections) ->
    lists:sort(fun(Conn1, Conn2) ->
                       Conn1#connection.ref < Conn2#connection.ref
               end,
               Connections).

%%----------------------------------------------------------------------
%% Changing options

do_change_options(Relay, NewOptions) ->
    RuntimOptions = add_announcement_option(NewOptions, Relay#relay.ref),
    trelay:change_options(Relay#relay.pid, RuntimOptions),
    lists:foreach(fun(Conn) ->
                          MasterPid = Conn#connection.pid,
                          trelay:change_options(MasterPid, RuntimOptions)
                  end,
                  Relay#relay.connections),
    Relay#relay{options = NewOptions}.


%%----------------------------------------------------------------------
%% Creating a list of relays

export_relays(TableId) ->
    L = lists:map(fun(Relay) ->
                          Ref = Relay#relay.ref,
                          Proto = Relay#relay.proto,
                          From = Relay#relay.from,
                          To = Relay#relay.to,
                          Options = Relay#relay.options,
                          Conns = export_connections(Relay#relay.connections),
                          SortedConns = sort_exported_connections(Conns),
                          {Ref, Proto, From, To, SortedConns, Options}
              end,
              ets:tab2list(TableId)),
    sort_exported_relays(L).
export_connections(ConnectionList) ->
    lists:map(fun(Connection) ->
                      RAddr = Connection#connection.r_addr,
                      RPort = Connection#connection.r_port,
                      Since = Connection#connection.since,
                      Ref = Connection#connection.ref,
                      {RAddr, RPort, Since, Ref}
              end,
              ConnectionList).

sort_exported_relays(Relays) ->
    lists:sort(fun(ExportRelay1, ExportRelay2) ->
                       Ref1 = element(1, ExportRelay1),
                       Ref2 = element(1, ExportRelay2),
                       Ref1 < Ref2
               end,
               Relays).

sort_exported_connections(Connections) ->
    lists:sort(fun(ExportConn1, ExportConn2) ->
                       Ref1 = element(4, ExportConn1),
                       Ref2 = element(4, ExportConn2),
                       Ref1 < Ref2
               end,
               Connections).


%%----------------------------------------------------------------------
%% Announcements

add_announcement_option(Options, Ref) ->
    Options ++ [{announce_to, {self(), Ref}}].

handle_announcement(RelayRef,
                    {new_connection, MasterPid, ConnRef, RemoteSide},
                    TableId) ->
    {RAddr, RPort} = RemoteSide,
    case find_relay(TableId, RelayRef) of
        {ok, Relay} ->
            NewConn = #connection{r_addr = RAddr,
                                  r_port = RPort,
                                  since = {date(), time()},
                                  pid = MasterPid,
                                  ref = ConnRef},
            NewConns = Relay#relay.connections ++ [NewConn],
            NewRelay = Relay#relay{connections = NewConns},
            ets:insert(TableId, NewRelay);
        _ ->
            error
    end;

handle_announcement(RelayRef, {closed_connection, ConnRef}, TableId) ->
    case find_relay(TableId, RelayRef) of
        {ok, Relay} ->
            NewConns = [Conn || Conn <- Relay#relay.connections,
                                Conn#connection.ref /= ConnRef],
            NewRelay = Relay#relay{connections = NewConns},
            ets:insert(TableId, NewRelay);
        _ ->
            error
    end.


%%----------------------------------------------------------------------
%% File handling: load


%% Perform conversions if necessary...
parse_relays({file_format_version, 1}, RelayTerms) ->
    %% This is version 1. No conversion is needed
    RelayTerms.

extract_new_relays(LoadedRelays, RunningRelays) ->
    lists:filter(fun(LoadedRelay) ->
                         not is_running_relay(LoadedRelay, RunningRelays)
                 end,
                 LoadedRelays).

is_running_relay(LoadedRelay, RunningRelays) ->
    lists:any(fun(RunningRelay) ->
                      is_same_relay(LoadedRelay, RunningRelay)
              end,
              RunningRelays).

is_same_relay(LoadedRelay, RunningRelay) ->
    (LoadedRelay#file_relay.proto == RunningRelay#relay.proto) and
        is_same_addr(LoadedRelay#file_relay.from, RunningRelay#relay.from).


is_same_addr({any, PortNumber}, {_AnyHostNameOrIp, PortNumber}) ->
    true;
is_same_addr({_AnyHostNameOrIp, PortNumber}, {any, PortNumber}) ->
    true;
is_same_addr({HostNameOrIp1, PortNumber}, {HostNameOrIp2, PortNumber}) ->
    IP1 = inet:getaddr(HostNameOrIp1, inet),
    IP2 = inet:getaddr(HostNameOrIp2, inet),
    IP1 == IP2;
is_same_addr({_AnyHost1, _PortNumber1}, {_AnyHost2, _PortNumber2}) ->
    false.



start_relays(TableId, FirstRef, NewFileRelays) ->
    lists:foldl(fun(FileRelay, Ref) ->
                        Proto   = FileRelay#file_relay.proto,
                        From    = FileRelay#file_relay.from,
                        To      = FileRelay#file_relay.to,
                        Options = FileRelay#file_relay.options,
                        case start_relay(Ref, Proto, From, To, Options) of
                            {ok, NewRelay} ->
                                ets:insert(TableId, NewRelay),
                                Ref + 1;
                            {error, _Reason} ->
                                Ref
                        end
                end,
                FirstRef,
                NewFileRelays).


%%----------------------------------------------------------------------
%% File handling: save

to_file_relays(TableId) ->
    Relays = ets:tab2list(TableId),
    RelaysSConns = lists:map(fun(Relay) ->
                                     Conns = Relay#relay.connections,
                                     SortedConns = sort_connections(Conns),
                                     Relay#relay{connections = SortedConns}
                             end,
                             Relays),
    SortedRelays = sort_relays(RelaysSConns),
    lists:map(fun(Relay) ->
                      #file_relay{proto    = Relay#relay.proto,
                                  from     = Relay#relay.from,
                                  to       = Relay#relay.to,
                                  options  = Relay#relay.options}
              end,
              SortedRelays).
